use std::{cmp::Ordering, iter};

use anyhow::Result;
use itertools::Itertools;
use rand::{seq::SliceRandom, thread_rng};
use rayon::iter::{ParallelBridge, ParallelIterator};

use super::{Agent, Chase, MaximizerAgent, MinimizerAgent};
use crate::{
	agents::{util::best_board_noop, Diffuse},
	board::{Board, PlayerAction},
	util::{Player, Symb},
};

pub struct Brutus {
	player: Player,
}

impl Brutus {
	pub fn new(player: Player) -> Self {
		Self { player }
	}

	fn step(&mut self, board: &Board, minimize: bool) -> Result<PlayerAction> {
		let symbols = [Symb::Minus, Symb::Times, Symb::Plus, Symb::Div]
			.into_iter()
			.filter(|x| !board.contains(*x))
			.collect_vec();

		if symbols.is_empty() {
			return if minimize {
				Chase::new(self.player).step_min(board)
			} else {
				Chase::new(self.player).step_max(board)
			};
		}

		// Number of free slots
		let n_free = board.get_board().iter().filter(|x| x.is_none()).count();

		// Number of slots we need to fill with numbers
		// Add one if we have two or fewer symbols available, so that we can
		// account for one unused symbol while keeping a reasonable runtime
		let n_fill = n_free - symbols.len() + if symbols.len() <= 2 { 1 } else { 0 };

		let mut items = iter::repeat(None)
			.take(n_fill)
			.chain(symbols.iter().map(|x| Some(x.clone())))
			.permutations(n_free)
			.unique()
			.par_bridge()
			.filter_map(move |x| {
				let mut tmp_board = board.clone();

				for (i, s) in x.iter().enumerate() {
					if let Some(s) = s {
						let pos = board.ith_empty_slot(i).unwrap();
						if !tmp_board.can_play(&PlayerAction { symb: *s, pos }) {
							return None;
						}
						tmp_board.get_board_mut()[pos] = Some(*s);
					}
				}

				let min = best_board_noop(&tmp_board, true);
				let max = best_board_noop(&tmp_board, false);
				if min.is_none() || max.is_none() {
					return None;
				}

				let v_min = min.unwrap().evaluate().unwrap();
				let v_max = max.unwrap().evaluate().unwrap();
				let v = v_min + v_max / 2.0;

				Some((tmp_board, v))
			})
			.collect::<Vec<_>>();

		if minimize {
			// Sort from smallest to biggest midpoint
			items.sort_by(|a, b| a.1.partial_cmp(&b.1).unwrap_or(Ordering::Equal));
		} else {
			items.sort_by(|a, b| b.1.partial_cmp(&a.1).unwrap_or(Ordering::Equal));
		}

		// TODO: why can `items` be empty?
		// We shouldn't need this escape hatch
		if items.is_empty() {
			return if minimize {
				Diffuse::new(self.player).step_min(board)
			} else {
				Diffuse::new(self.player).step_max(board)
			};
		}

		let (t, _) = items.first().unwrap();

		let mut symbols = symbols.clone();
		symbols.shuffle(&mut thread_rng());

		// Place a random unused symbol
		for target_s in symbols {
			for (i, s) in t.get_board().iter().enumerate() {
				if let Some(s) = s {
					if board.get_board()[i].is_none() && target_s == *s {
						return Ok(PlayerAction { pos: i, symb: *s });
					}
				}
			}
		}

		// Final escape hatch, if we didn't decide to place any symbols
		// (which is possible, since we add one to free_spots above!)
		if minimize {
			Chase::new(self.player).step_min(board)
		} else {
			Chase::new(self.player).step_max(board)
		}
	}
}

impl Agent for Brutus {
	fn name(&self) -> &'static str {
		"Brutus"
	}

	fn player(&self) -> Player {
		self.player
	}
}

impl MinimizerAgent for Brutus {
	fn step_min(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, true)
	}
}

impl MaximizerAgent for Brutus {
	fn step_max(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, false)
	}
}
