use anyhow::{bail, Result};
use std::num::NonZeroU8;

use super::{Agent, MaximizerAgent, MinimizerAgent, Random};
use crate::{
	agents::util::free_slots_by_influence,
	board::{Board, PlayerAction},
	util::{Player, Symb},
};

pub struct Chase {
	player: Player,
}

impl Chase {
	pub fn new(player: Player) -> Self {
		Self { player }
	}

	fn step(&mut self, board: &Board, minimize: bool) -> Result<PlayerAction> {
		let available_numbers = (0..=9)
			.map(|x| match x {
				0 => Symb::Zero,
				x => Symb::Number(NonZeroU8::new(x).unwrap()),
			})
			.filter(|x| !board.contains(*x))
			.collect::<Vec<_>>();

		// For the code below, we must guarantee that
		// min_slots + max_slots <= available_numbers.len
		let n_free = board.get_board().iter().filter(|x| x.is_none()).count();
		if available_numbers.len() < n_free || n_free >= 10 {
			return Random::new(self.player).step_min(board);
		}

		let t = free_slots_by_influence(board);
		if t.is_none() {
			bail!("could not compute next move!")
		}
		let t = t.unwrap();

		if t.is_empty() {
			return Random::new(self.player).step_min(board);
		}

		let (pos, val) = t[0];

		// Choose next number if we can't make the a move.
		// Prevents division by zero.
		// This isn't perfect, and may fail if we run out of numbers
		// (This is, however, very unlikely)
		let mut symb = None;
		let mut offset = 0;
		while symb.is_none()
			|| !board.can_play(&PlayerAction {
				symb: symb.unwrap(),
				pos,
			}) {
			symb = Some({
				if minimize {
					if val >= 0.0 {
						available_numbers[offset]
					} else {
						available_numbers[available_numbers.len() - 1 - offset]
					}
				} else if val <= 0.0 {
					available_numbers[offset]
				} else {
					available_numbers[available_numbers.len() - 1 - offset]
				}
			});
			offset += 1;
		}

		Ok(PlayerAction {
			symb: symb.unwrap(),
			pos,
		})
	}
}

impl Agent for Chase {
	fn name(&self) -> &'static str {
		"Chase"
	}

	fn player(&self) -> Player {
		self.player
	}
}

impl MinimizerAgent for Chase {
	fn step_min(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, true)
	}
}

impl MaximizerAgent for Chase {
	fn step_max(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, false)
	}
}
