use std::fmt::Display;

use clap::{Parser, ValueEnum};

use crate::{
	agents::{Brutus, Chase, Diffuse, Human, MaximizerAgent, MinimizerAgent, Random},
	util::Player,
};

#[derive(Parser, Debug)]
#[command(about)]
pub struct Cli {
	/// The agent that controls the Blue (opponent) player
	pub blue: AgentSelector,

	/// The agent that controls the Red (home) player
	#[arg(long, default_value = "human")]
	pub red: AgentSelector,

	/// If this is greater than one, repeat the game this many times and print a summary.
	#[arg(long, short, default_value = "0")]
	pub repeat: usize,
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, PartialOrd, Ord, ValueEnum)]
pub enum AgentSelector {
	/// A human agent. Asks for input.
	Human,

	/// A random agent (very easy)
	Random,

	/// A simple extremum-chaser (easy)
	Chase,

	/// A smarter extremum-chaser (medium)
	Diffuse,

	/// A very smart brute-force agent (hard)
	Brutus,
}

impl AgentSelector {
	pub fn get_maximizer(&self, player: Player) -> Box<dyn MaximizerAgent> {
		match self {
			Self::Random => Box::new(Random::new(player)),
			Self::Chase => Box::new(Chase::new(player)),
			Self::Diffuse => Box::new(Diffuse::new(player)),
			Self::Brutus => Box::new(Brutus::new(player)),
			Self::Human => Box::new(Human::new(player)),
		}
	}

	pub fn get_minimizer(&self, player: Player) -> Box<dyn MinimizerAgent> {
		match self {
			Self::Random => Box::new(Random::new(player)),
			Self::Chase => Box::new(Chase::new(player)),
			Self::Diffuse => Box::new(Diffuse::new(player)),
			Self::Brutus => Box::new(Brutus::new(player)),
			Self::Human => Box::new(Human::new(player)),
		}
	}
}

impl Display for AgentSelector {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		write!(
			f,
			"{}",
			match self {
				Self::Random => "random",
				Self::Diffuse => "diffuse",
				Self::Brutus => "brutus",
				Self::Chase => "chase",
				Self::Human => "human",
			}
		)
	}
}
