use std::{
	cmp::Ordering,
	io::{stdout, Write},
};

use anyhow::{bail, Result};
use clap::Parser;
use rayon::iter::{IntoParallelIterator, ParallelIterator};
use termion::color::{self};

mod agents;
mod board;
mod cli;
mod util;
use board::Board;
use util::{Player, Symb};

fn play_silent(
	maxi: &mut dyn agents::MaximizerAgent,
	mini: &mut dyn agents::MinimizerAgent,
) -> Result<Board> {
	let mut board = Board::new();
	let mut is_maxi_turn = true;

	while !board.is_done() {
		// Take action
		let action = if is_maxi_turn {
			maxi.step_max(&board)?
		} else {
			mini.step_min(&board)?
		};

		if !board.play(
			action,
			if is_maxi_turn {
				maxi.player()
			} else {
				mini.player()
			},
		) {
			bail!(
				"agent {} made an invalid move {}!",
				if is_maxi_turn {
					maxi.name()
				} else {
					mini.name()
				},
				action
			)
		}

		is_maxi_turn = !is_maxi_turn;
	}

	Ok(board)
}

fn play(
	maxi: &mut dyn agents::MaximizerAgent,
	mini: &mut dyn agents::MinimizerAgent,
) -> Result<Board> {
	let mut board = Board::new();
	let mut is_first_turn = true;
	let mut is_maxi_turn = true;

	let board_label = format!(
		"{}{:<6}{}",
		color::Fg(color::LightBlack),
		maxi.player().to_string(),
		color::Fg(color::Reset)
	);

	while !board.is_done() {
		// Print board
		println!(
			"\r{}{}{}{}",
			board_label,
			if is_first_turn { '╓' } else { '║' },
			board.prettyprint()?,
			if is_first_turn { '╖' } else { '║' },
		);
		is_first_turn = false;

		print!("Thinking...");
		stdout().flush()?;

		// Take action
		let action = if is_maxi_turn {
			maxi.step_max(&board)?
		} else {
			mini.step_min(&board)?
		};

		if !board.play(
			action,
			if is_maxi_turn {
				maxi.player()
			} else {
				mini.player()
			},
		) {
			bail!(
				"agent {} made an invalid move {}!",
				if is_maxi_turn {
					maxi.name()
				} else {
					mini.name()
				},
				action
			)
		}

		is_maxi_turn = !is_maxi_turn;
	}

	println!("\r{}║{}║", board_label, board.prettyprint()?);
	println!("\r{}╙{}╜", board_label, " ".repeat(board.size()));
	Ok(board)
}

fn main() -> Result<()> {
	let cli = cli::Cli::parse();

	rayon::ThreadPoolBuilder::new()
		.num_threads(8)
		.build_global()
		.unwrap();

	if cli.repeat > 1 {
		let x = (0..cli.repeat)
			.into_par_iter()
			.filter_map(|_| {
				let mut maxi = cli.red.get_maximizer(Player::Red);
				let mut mini = cli.blue.get_minimizer(Player::Blue);
				let red_board = match play_silent(&mut *maxi, &mut *mini) {
					Ok(x) => x,
					Err(e) => {
						println!("Error: {e}");
						return None;
					}
				};

				let mut maxi = cli.blue.get_maximizer(Player::Blue);
				let mut mini = cli.red.get_minimizer(Player::Red);
				let blue_board = match play_silent(&mut *maxi, &mut *mini) {
					Ok(x) => x,
					Err(e) => {
						println!("Error: {e}");
						return None;
					}
				};

				match red_board.evaluate().partial_cmp(&blue_board.evaluate()) {
					Some(Ordering::Equal) => None,
					Some(Ordering::Greater) => Some(Player::Red),
					Some(Ordering::Less) => Some(Player::Blue),
					None => {
						println!("Error");
						None
					}
				}
			})
			.collect::<Vec<_>>();

		let red_wins = x.iter().filter(|x| **x == Player::Red).count();
		let blue_wins = x.iter().filter(|x| **x == Player::Blue).count();

		println!("Played {} rounds\n", cli.repeat);
		println!(
			"Red  win rate: {:.2}  ({})",
			red_wins as f32 / cli.repeat as f32,
			cli.red,
		);
		println!(
			"Blue win rate: {:.2}  ({}) ",
			blue_wins as f32 / cli.repeat as f32,
			cli.blue,
		);
	} else {
		let mut maxi = cli.red.get_maximizer(Player::Red);
		let mut mini = cli.blue.get_minimizer(Player::Blue);
		let red_board = play(&mut *maxi, &mut *mini)?;

		if red_board.is_done() {
			println!(
				"\r\n{}{} score:{} {:.2}\n\n",
				color::Fg(maxi.player().color()),
				maxi.name(),
				color::Fg(color::Reset),
				red_board.evaluate().unwrap()
			);
		} else {
			println!(
				"\r\n{}Quitting{}\r\n",
				color::Fg(color::Red),
				color::Fg(color::Reset),
			);
			return Ok(());
		}

		let mut maxi = cli.blue.get_maximizer(Player::Blue);
		let mut mini = cli.red.get_minimizer(Player::Red);
		let blue_board = play(&mut *maxi, &mut *mini)?;

		if blue_board.is_done() {
			println!(
				"\r\n{}{} score:{} {:.2}\n\n",
				color::Fg(maxi.player().color()),
				maxi.name(),
				color::Fg(color::Reset),
				blue_board.evaluate().unwrap()
			);
		} else {
			println!(
				"\r\n{}Quitting{}\r\n",
				color::Fg(color::Red),
				color::Fg(color::Reset),
			);
			return Ok(());
		}

		match red_board.evaluate().partial_cmp(&blue_board.evaluate()) {
			Some(Ordering::Equal) => {
				println!("\r\nTie");
			}
			Some(Ordering::Greater) => {
				println!(
					"\r\n{}Red ({}){} wins!",
					color::Fg(Player::Red.color()),
					cli.red,
					color::Fg(color::Reset),
				);
			}
			Some(Ordering::Less) => {
				println!(
					"\r\n{}Blue ({}){} wins!",
					color::Fg(Player::Blue.color()),
					cli.blue,
					color::Fg(color::Reset),
				);
			}
			None => {
				println!(
					"\r\n{}Error{}",
					color::Fg(color::Red),
					color::Fg(color::Reset),
				);
			}
		}
	}

	Ok(())
}
