use std::{
	fmt::{Debug, Display},
	num::NonZeroU8,
};

use termion::color::{self, Color};

#[derive(PartialEq, Eq, Clone, Copy)]
pub enum Player {
	Red,
	Blue,
}

impl Player {
	pub fn color(&self) -> &dyn Color {
		match self {
			Player::Red => &color::LightRed,
			Player::Blue => &color::LightBlue,
		}
	}
}

impl Display for Player {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		let s = match self {
			Self::Red => "Red ",
			Self::Blue => "Blue",
		};
		write!(f, "{s}")
	}
}

#[derive(PartialEq, Eq, Clone, Copy, Hash)]
pub enum Symb {
	Number(NonZeroU8),
	Zero,
	Plus,
	Minus,
	Times,
	Div,
}

impl Display for Symb {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		match self {
			Self::Number(x) => write!(f, "{x}")?,
			Self::Zero => write!(f, "0")?,
			Self::Plus => write!(f, "+")?,
			Self::Minus => write!(f, "-")?,
			Self::Div => write!(f, "÷")?,
			Self::Times => write!(f, "×")?,
		}
		Ok(())
	}
}
impl Debug for Symb {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		Display::fmt(self, f)
	}
}

impl Symb {
	/// Is this symbol a plain binary operator?
	pub fn is_op(&self) -> bool {
		matches!(self, Symb::Div | Symb::Plus | Symb::Times | Symb::Minus)
	}

	pub fn is_minus(&self) -> bool {
		self == &Self::Minus
	}

	pub const fn get_char(&self) -> Option<char> {
		match self {
			Self::Plus => Some('+'),
			Self::Minus => Some('-'),
			Self::Times => Some('×'),
			Self::Div => Some('÷'),
			Self::Zero => Some('0'),
			Self::Number(x) => match x.get() {
				1 => Some('1'),
				2 => Some('2'),
				3 => Some('3'),
				4 => Some('4'),
				5 => Some('5'),
				6 => Some('6'),
				7 => Some('7'),
				8 => Some('8'),
				9 => Some('9'),
				_ => None,
			},
		}
	}

	pub const fn from_char(c: char) -> Option<Self> {
		match c {
			'1' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(1) })),
			'2' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(2) })),
			'3' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(3) })),
			'4' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(4) })),
			'5' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(5) })),
			'6' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(6) })),
			'7' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(7) })),
			'8' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(8) })),
			'9' => Some(Self::Number(unsafe { NonZeroU8::new_unchecked(9) })),
			'0' => Some(Self::Zero),
			'+' => Some(Self::Plus),
			'-' => Some(Self::Minus),
			'*' => Some(Self::Times),
			'/' => Some(Self::Div),
			'×' => Some(Self::Times),
			'÷' => Some(Self::Div),
			_ => None,
		}
	}
}
